const Dexie = require('dexie');

let db = null;
async function initDatabase() {
  try {
    db = new Dexie('StubsDatabase2');
    db.version(1).stores({
      chatgpt_requests: '++id, user_id, request_type, current_task, full_prompt, assistant_message, ret_data, request_date, llm_provider, model',
      chatgpt_responses: '++id, user_id, request_type, current_task, full_prompt, assistant_message, ret_data, request_date, uid, llm_provider, model, input_tokens, output_tokens',
      dependency_graph: '++id, project, caller, line_caller, callee, line_callee, type, name, parent_function',
      dependency_groups: '++id, project, group_index, file_paths',
      file_descriptions: '++id, project_name, file_path, description, update_date, md5, last_modified_date',
      embeddings: '++id, project_name, vector_data, update_date',
      documents: '++id, project_name, document_name, original_url, url, vector_data, markdown, update_date, active, status',
      stub_files: '++id, md5, line_count, needs_stub, is_included, full_path, file_path, filename, project_name, stub_file',
      debug_data: '++id, project, message_type, message_body, error, stack_trace, timestamp',
      apply_changes_log: '++id, all_relevant_text, assistant_message, final_original_paths, final_short_paths, just_files_linux, my_project_data, files_snapshot'
    });
    await db.open();
    console.log("Database opened successfully");
  } catch (error) {
    console.error("Error opening database:", error);
    console.error("Error stack:", error.stack);
  }
};

async function getMockResponse(requestType, currentTask, llmProvider) {
  try {
    const rows = await db.chatgpt_responses
      .where({request_type: requestType, current_task: currentTask, llm_provider: llmProvider})
      .reverse()
      .limit(1)
      .toArray();

    if (rows.length === 0) {
      return null;
    }

    const row = rows[0];
    const retData = row.ret_data;

    if (!retData || retData.length === 0) {
      return null;
    }

    let parsedData;
    try {
      parsedData = JSON.parse(retData);
    } catch (error) {
      return null;
    }

    parsedData.mock = 'yes';
    const retString = JSON.stringify(parsedData);

    return retString;
  } catch (err) {
    console.error('Error getting mock response:', err);
    return null;
  }
}

async function logRequest(userId, requestType, currentTask, fullPrompt, assistantMessage, retData, llmProvider, model) {
  const now = new Date();
  //const requestDate = now.toISOString().slice(0, 19).replace('T', ' ');
  console.log("logging request");

  try {
    await db.chatgpt_requests.add({
      user_id: userId,
      request_type: requestType,
      current_task: currentTask,
      full_prompt: fullPrompt,
      assistant_message: assistantMessage,
      ret_data: retData,
      request_date: now,
      llm_provider: llmProvider,
      model: model
    });
  } catch (err) {
    console.error('Error logging request:', err);
  }
}

async function logResponse(userId, requestType, currentTask, fullPrompt, assistantMessage, retData, uid, llmProvider, model, inputTokens, outputTokens) {
  const now = new Date();
  const requestDate = now.toISOString().slice(0, 19).replace('T', ' ');

  let modelFirst = model;
  if(model){
      modelFirst = model.split('|')[0];
  }
  try {
    await db.chatgpt_responses.add({
      user_id: userId,
      request_type: requestType,
      current_task: currentTask,
      full_prompt: fullPrompt,
      assistant_message: assistantMessage,
      ret_data: retData,
      request_date: requestDate,
      uid: uid,
      llm_provider: llmProvider,
      model: modelFirst,
      input_tokens: inputTokens,
      output_tokens: outputTokens
    });
  } catch (err) {
    console.error('Error logging response:', err);
  }
}

async function saveDependencyGraph(projectName, dependencyGraph) {
  try {
    // Clear out the existing rows for the project
    await db.dependency_graph.where('project').equals(projectName).delete();
    
    await db.transaction('rw', db.dependency_graph, async () => {
      for (const dependency of dependencyGraph) {
        await db.dependency_graph.add({
          project: projectName,
          caller: dependency.caller,
          line_caller: dependency.line_caller,
          callee: dependency.callee,
          line_callee: dependency.line_callee,
          type: dependency.type,
          name: dependency.name,
          parent_function: dependency.parent_function
        });
      }
    });
    console.log('Dependency graph saved to Dexie');
  } catch (err) {
    console.error('Error saving dependency graph:', err);
  }
}
async function saveDependencyGroups(projectName, dependencyGroups) {
  try {

    // Clear out the existing rows for the project
    await db.dependency_groups.where('project').equals(projectName).delete();

    await db.transaction('rw', db.dependency_groups, async () => {
      for (let i = 0; i < dependencyGroups.length; i++) {
        const group = dependencyGroups[i];
        await db.dependency_groups.add({
          project: projectName,
          group_index: i,
          file_paths: JSON.stringify(group)
        });
      }
    });
    console.log('Dependency groups saved to Dexie');
  } catch (err) {
    console.error('Error saving dependency groups:', err);
  }
}

async function getDependencyGraph(projectName) {
  try {
    const dependencyGraph = await db.dependency_graph
      .where('project')
      .equals(projectName)
      .toArray();

    return dependencyGraph;
  } catch (err) {
    console.error('Error getting dependency graph:', err);
    return [];
  }
}

async function getFileNamesFromDependencyGraph(projectName) {
  try {
    const fileNames = await db.dependency_graph
      .where('project')
      .equals(projectName)
      .toArray(row => row.caller);

    return [...new Set(fileNames)];
  } catch (err) {
    console.error('Error getting file names from dependency graph:', err);
    return [];
  }
}


async function getDependencyGroups(projectName) {
  try {
    const dependencyGroups = await db.dependency_groups
      .where('project')
      .equals(projectName)
      .toArray();

    // Parse the file paths from JSON string to array
    const parsedGroups = dependencyGroups.map(group => ({
      ...group,
      file_paths: JSON.parse(group.file_paths)
    }));

    return parsedGroups;
  } catch (err) {
    console.error('Error getting dependency groups:', err);
    return [];
  }
}

async function getHistoryData() {
  try {
    const historyData = await db.chatgpt_responses
      .where('request_type')
      .equals('do_new_programming_task')
      .reverse() 
      .toArray();

    return historyData;
  } catch (err) {
    console.error('Error getting history data:', err);
    return [];
  }
}

async function getAllHistoryData() {
  try {
    const historyData = (await db.chatgpt_responses.toArray()).slice().reverse();

    // Group responses by current_task
    const groupedData = historyData.reduce((acc, response) => {
      if (!acc[response.current_task]) {
        // Replace the space between date and time with 'T' and add 'Z' at the end to indicate UTC time
        console.log(response.request_date);
        let localDate = "???";
        if(response.request_date.toString().indexOf('GMT') >= 0){
            console.log("alt date");
            localDate = response.request_date.toString();
        }else{
            let dateStringUTC = response.request_date.replace(' ', 'T') + 'Z';
            let date = new Date(dateStringUTC);

            
            if (isNaN(date.getTime())) {
              console.error('Invalid Date:', response.request_date);
            }else{                 
                let options = { hour: 'numeric', minute: 'numeric', hour12: true };
                localDate = date.toLocaleString('en-US', options).toString().replace(' ', '').toLowerCase();
                console.log("Local Date: " );
                console.log(localDate);        
            }            
        }
        acc[response.current_task] = {
          current_task: response.current_task,
          formatted_date: localDate,
          totalCost: 0,
          isProgrammingTask: false
        };
      }
      
      const retData = JSON.parse(response.ret_data);
      let inputTokens = response.input_tokens;
      let outputTokens = response.output_tokens;
      if(!inputTokens || inputTokens <= 0){
          inputTokens = COST.estimateTokens(response.full_prompt);
      }
      if(!outputTokens || outputTokens <= 0){
          outputTokens = COST.estimateTokens(retData.orig_response);
      }      
      
      const cost = COST.getTokenCost(response.model, inputTokens, outputTokens);
      acc[response.current_task].totalCost += cost;

      if (response.request_type === 'do_new_programming_task') {
        acc[response.current_task].isProgrammingTask = true;
      }
      
      // Format totalCost to two decimal places
      acc[response.current_task].totalCost = Number(acc[response.current_task].totalCost.toFixed(2));

      return acc;
    }, {});
    

    const fullHistoryData = historyData.map(response => {
      const retData = JSON.parse(response.ret_data);
      let inputTokens = response.input_tokens;
      let outputTokens = response.output_tokens;
      if(!inputTokens || inputTokens <= 0){
          inputTokens = COST.estimateTokens(response.full_prompt);
      }
      if(!outputTokens || outputTokens <= 0){
          outputTokens = COST.estimateTokens(retData.orig_response);
      }
      
      
      const cost = COST.getTokenCost(response.model, inputTokens, outputTokens);
      return {
        ...response,
        totalCost: Number(cost.toFixed(2)),
        input_tokens: inputTokens,
        output_tokens: outputTokens
      };
    });    

    return [Object.values(groupedData).filter(task => task.isProgrammingTask), fullHistoryData];
  } catch (err) {
    console.error('Error getting history data:', err);
    return [[], []];
  }
}


async function getTotalUsedTokens(model, start_date = null, end_date = null) {
  console.log("getting cost for " + model);

  let query = db.chatgpt_responses
    .where('model')
    .equals(model);
  
  if (start_date) {
    query = query.and(response => new Date(response.request_date) >= new Date(start_date));
  }

  if (end_date) {
    query = query.and(response => new Date(response.request_date) <= new Date(end_date));
  }
  
  const responses = await query.toArray();

  let totalPromptTokens = 0;
  let totalCompletionTokens = 0;

  responses.forEach(response => {
    //console.log("cost record");
    //console.log(retData);
    //console.log("Request Date:", response.request_date);
    //console.log("Input Tokens:", retData.input_tokens);
    //console.log("Output Tokens:", retData.output_tokens);
    //console.log("--------------------");
    
    totalPromptTokens += response.input_tokens || 0;
    totalCompletionTokens += response.output_tokens || 0;
  });

  return {
    input_tokens: totalPromptTokens,
    output_tokens: totalCompletionTokens
  };
}

async function saveFileDescriptions(projectName, fileDescriptions) {
  try {
    await db.transaction('rw', db.file_descriptions, async () => {
      // Get all existing records for the project
      const existingRecords = await db.file_descriptions
        .where('project_name')
        .equals(projectName)
        .toArray();

      console.log("existing records");
      console.log(existingRecords);

      // Create a map of existing file paths for quick lookup
      const existingFilePaths = new Set(existingRecords.map(record => record.file_path));

      for (const fileDesc of fileDescriptions) {
        console.log("seeing if " + fileDesc.file_path);
        console.log("is in set ");
        console.log(existingFilePaths);
          
        const md5 = crypto.createHash('md5').update(fileDesc.file_path).digest('hex');
        const updateDate = Date.now();
          
        if (existingFilePaths.has(fileDesc.file_path)) {
            console.log("modifying description of " + fileDesc.file_path);
          // If the file path exists, update the existing record
          await db.file_descriptions
            .where('project_name')
            .equals(projectName)
            .and(record => record.file_path === fileDesc.file_path)
            .modify({  description: fileDesc.description,
            	update_date: updateDate,
            	md5: md5
            });
        } else {
          // If the file path doesn't exist, insert a new record
            console.log("adding description of " + fileDesc.file_path);
          
          await db.file_descriptions.add({
            project_name: projectName,
            file_path: fileDesc.file_path,
            description: fileDesc.description
          });
        }
      }
    });
    console.log('File descriptions saved to Dexie');
  } catch (err) {
    console.error('Error saving file descriptions:');
    console.log(err);
  }
}

// Add these functions to save and retrieve embeddings from Dexie
async function saveEmbeddingsToDexie(projectName, embeddings, docfiles) {
  try {
    await db.transaction('rw', db.embeddings, async () => {
      // Clear existing embeddings for the project
      await db.embeddings.where('project_name').equals(projectName).delete();

      let updateDate = Date.now();
      // Add new embeddings
      for (let i = 0; i < embeddings.length; i++) {
        await db.embeddings.add({
          project_name: projectName,
          vector_data: embeddings[i],
          filename: docfiles[i],
          update_date: updateDate
        });
      }
    });
    console.log('Embeddings saved to Dexie');
  } catch (err) {
    console.error('Error saving embeddings:', err);
  }
}

async function getEmbeddingsFromDexie(projectName) {
  try {
    const embeddings = await db.embeddings
      .where('project_name')
      .equals(projectName)
      .toArray();

    return embeddings;
  } catch (err) {
    console.error('Error getting embeddings:', err);
    return [];
  }
}

async function deleteEmbedding(projectName, filePath) {
  try {
    await db.embeddings
      .where({project_name: projectName, filename: filePath})
      .delete();
    console.log(`Embedding deleted for project: ${projectName}, file: ${filePath}`);
  } catch (err) {
    console.error('Error deleting embedding:', err);
  }
}

async function getProjectFileDescriptions(projectName) {
  try {
    const fileDescriptions = await db.file_descriptions
      .where('project_name')
      .equals(projectName)
      .toArray();

    return fileDescriptions;
  } catch (err) {
    console.error('Error getting all file descriptions:', err);
    return [];
  }
}

async function saveStubFilesToDexie(projectName, stubFiles) {
  try {
    await db.transaction('rw', db.stub_files, async () => {
      // Clear existing stub files for the project
      await db.stub_files.where('project_name').equals(projectName).delete();

      // Add new stub files
      for (const file of stubFiles) {
        await db.stub_files.add({
          md5: file.md5,
          line_count: file.line_count,
          needs_stub: file.needs_stub,
          is_included: file.is_included,
          full_path: file.full_path,
          file_path: file.file_path,
          filename: file.filename,
          project_name: projectName,
          stub_file: file.stub_file
        });
      }
    });
    console.log('Stub files saved to Dexie for project ' + projectName);
  } catch (err) {
    console.error('Error saving stub files:', err);
  }
}

async function updateStubFileContent(projectName, filePath, stubFileContent, md5Hash) {
  try {
    const updatedCount = await db.stub_files
      .where({ project_name: projectName, file_path: filePath })
      .modify({ stub_file: stubFileContent, md5: md5Hash });

    if (updatedCount === 0) {
      console.error(`No stub file found for project "${projectName}" and file path "${filePath}"`);
    } else {
      console.log(`Stub file content updated for project "${projectName}" and file path "${filePath}"`);
    }
  } catch (err) {
    console.error('Error updating stub file content:', err);
  }
}

async function deleteStubFile(projectName, filePath) {
  try {
    const deletedCount = await db.stub_files
      .where({ project_name: projectName, file_path: filePath })
      .delete();

    if (deletedCount === 0) {
      console.error(`No stub file found for project "${projectName}" and file path "${filePath}"`);
    } else {
      console.log(`Stub file deleted for project "${projectName}" and file path "${filePath}"`);
    }
  } catch (err) {
    console.error('Error deleting stub file:', err);
  }
}

async function getStubFilesFromDexie(projectName) {  
  try {
    console.log(`Attempting to fetch stub files for project: ${projectName}`);
    const stubFiles = await db.stub_files
      .where('project_name')
      .equals(projectName)
      .toArray();
    console.log(`Retrieved ${stubFiles.length} stub files`);
    return stubFiles;
  } catch (err) {
    console.error('Error getting stub files:', err);
    console.error('Error stack:', err.stack);
    return [];
  }
}

async function getStubFileByPath(projectName, filePath) {
  try {
    const stubFile = await db.stub_files
      .where({ project_name: projectName, file_path: filePath })
      .first();

    return stubFile;
  } catch (err) {
    console.error('Error getting stub file by path:', err);
    return null;
  }
}

async function saveDebugData(debugData) {
  try {
    await db.debug_data.add({
      project: debugData.project,
      message_type: debugData.messageType,
      message_body: debugData.messageBody,
      error: debugData.error,
      stack_trace: debugData.stackTrace,
      timestamp: new Date()
    });
    console.log('Debug data saved to Dexie');
  } catch (err) {
    console.error('Error saving debug data:', err);
  }
}

async function printTableObjects(tableName, key, value) {
  try {
    const rows = await db[tableName].where(key).equals(value).toArray();
    if (rows.length === 0) {
      console.log(`No records found in table ${tableName} where ${key} equals ${value}`);
      return;
    }
    console.log(`Records in table ${tableName} where ${key} equals ${value}:`);
    rows.forEach(row => console.log(row));
  } catch (err) {
    console.error(`Error printing table objects for table ${tableName} where ${key} equals ${value}:`, err);
  }
}

async function truncateTable(tableName) {
  try {
    await db[tableName].clear();
    console.log(`Table ${tableName} truncated successfully.`);
  } catch (err) {
    console.error(`Error truncating table ${tableName}:`, err);
  }
}

async function clearDatabase() {
  try {
    await db.delete();
    console.log("Database deleted successfully.");
    await initDatabase();
    console.log("Database reinitialized successfully.");
  } catch (err) {
    console.error("Error clearing the database:", err);
  }
}

async function logApplyChanges(allRelevantText, assistantMessage, finalOriginalPaths, finalShortPaths, justFilesLinux, myProjectData, filesSnapshot) {
  try {
    await db.apply_changes_log.add({
      all_relevant_text: allRelevantText,
      assistant_message: assistantMessage,
      final_original_paths: finalOriginalPaths,
      final_short_paths: finalShortPaths,
      just_files_linux: justFilesLinux,
      my_project_data: myProjectData,
      files_snapshot: filesSnapshot
    });

    console.log('Apply changes logged to Dexie');
  } catch (err) {
    console.error('Error logging apply changes:', err);
  }
}

async function getDocumentationEmbeddings(projectName) {
  try {
    const documentationEmbeddings = await db.documentation_embeddings
      .where('project_name')
      .equals(projectName)
      .toArray();

    return documentationEmbeddings;
  } catch (err) {
    console.error('Error getting documentation embeddings:', err);
    return [];
  }
}

async function getAllRequestsByUid(uid) {
  try {
    const rows = await db.chatgpt_responses
      .where('uid')
      .equals(uid)
      .toArray();

    if (rows.length === 0) {
      return null;
    }
    return rows;
  } catch (err) {
    console.error('Error getting task by request type and uid:', err);
    return null;
  }
}
async function getLatestResponseByRequestType(requestType) {
  try {
    const rows = await db.chatgpt_responses
      .where('request_type')
      .equals(requestType)
      .reverse()
      .limit(1)
      .toArray();

    if (rows.length === 0) {
      return null;
    }

    return rows[0];
  } catch (err) {
    console.error('Error getting latest task by request type:', err);
    return null;
  }
}

async function getDocuments(projectName) {
  try {
    const documents = await db.documents
      .where('project_name')
      .equals(projectName)
      .toArray();

    return documents;
  } catch (err) {
    console.error('Error getting documents:', err);
    return [];
  }
}

async function saveDocument(projectName, documentName, originalUrl, url, vectorData, status, crawlType, markdown) {
  try {
    await db.documents.add({
      project_name: projectName,
      original_url: originalUrl,
      url: url,
      document_name: documentName,
      vector_data: vectorData,
      update_date: new Date(),
      active: true,
      status: status,
      crawl_type: crawlType,
      markdown: markdown
    });
    console.log('Document saved to Dexie');
  } catch (err) {
    console.error('Error saving document:', err);
  }  
}

async function getDocumentByName(projectName, docName) {
    try {
        const document = await db.documents
            .where({project_name: projectName, document_name: docName})
            .first();
        return document;
    } catch (err) {
        console.error('Error getting document by name:', err);
        return null;
    }
}
async function updateDocument(projectName, documentName, updateData) {
    try {
        await db.documents
            .where({project_name: projectName, document_name: documentName})
            .modify(document => {
                Object.assign(document, updateData);
                document.update_date = new Date();
            });
        console.log('Document updated in Dexie');
    } catch (err) {
        console.error('Error updating document:', err);
    }
}
/*
async function updateDocument(projectName, embeddingName, crawledData, active) {
    try {
        await db.documents
            .where({project_name: projectName, embedding_name: embeddingName})
            .modify(document => {
                document.url = crawledData[0].url;
                document.document_name = crawledData[0].original_url;
                document.vector_data = JSON.stringify(crawledData);
                document.update_date = new Date();
                document.active = active;
            });
        console.log('Document updated in Dexie');
    } catch (err) {
        console.error('Error updating document:', err);
    }
}
*/
async function setDocumentActive(projectName, embeddingName, isActive) {
  try {
    await db.documents
      .where({ project_name: projectName, embedding_name: embeddingName })
      .modify({ active: isActive });
    console.log(`Document ${embeddingName} for project ${projectName} set to ${isActive ? 'active' : 'inactive'}`);
  } catch (err) {
    console.error('Error setting document active status:', err);
  }
}

async function deleteDocument(projectName, documentName) {
  try {
    await db.documents
      .where({ project_name: projectName, document_name: documentName })
      .delete();
    console.log(`Document ${documentName} deleted from project ${projectName}`);
  } catch (err) {
    console.error('Error deleting document:', err);
  }
}

module.exports = {
  initDatabase,
  logRequest,
  logResponse,
  getMockResponse,
  saveDependencyGraph,
  saveDependencyGroups,
  getDependencyGraph,
  getDependencyGroups,
  saveFileDescriptions,
  saveEmbeddingsToDexie,
  getProjectFileDescriptions,
  getEmbeddingsFromDexie,
  deleteEmbedding,
  saveStubFilesToDexie,
  updateStubFileContent,
  deleteStubFile,
  getStubFilesFromDexie,
  getStubFileByPath,
  getHistoryData,
  getAllHistoryData,
  getTotalUsedTokens,
  truncateTable,
  printTableObjects,
  saveDebugData,
  clearDatabase,
  logApplyChanges,
  getAllRequestsByUid,
  getLatestResponseByRequestType,
  getDocuments,
  saveDocument,
  getDocumentByName,
  updateDocument,
  setDocumentActive,
  deleteDocument
};
