const mysql = require('mysql');
const util = require('util');
let pool = null;

async function initDatabase(mysqlConfig){
    //console.log("mysql config");
    //console.log(mysqlConfig);
    if(pool == null){
        pool = mysql.createPool({
          connectionLimit: 10,
          host: mysqlConfig['server'],
          user: mysqlConfig['user'],
          password: mysqlConfig['password'],
          database: mysqlConfig['name']
        });

        pool.query = util.promisify(pool.query);
    }
    console.log("done init");
}

async function getMockResponse(requestType, currentTask, llmProvider) {
    console.log("requestType: " + requestType);
    console.log("currentTask: " + currentTask);
    console.log("llmProvider: " + llmProvider);
    
  try {
    const rows = await pool.query('SELECT * FROM chatgpt_responses WHERE request_type = ? AND current_task = ? AND llm_provider = ? ORDER BY id DESC LIMIT 1', 
        [requestType, currentTask, llmProvider]);

    if (rows.length === 0) {
      return null;
    }

    const row = rows[0];
    const retData = row.ret_data;

    if (!retData || retData.length === 0) {
      return null;
    }

    let parsedData;
    try {
      parsedData = JSON.parse(retData);
    } catch (error) {
      return null;
    }

    parsedData.mock = 'yes';
    //const retString = JSON.stringify(parsedData);

    return parsedData;
  } catch (err) {
    console.error('Error getting mock response:', err);
    return null;
  }
}

async function getMockByFullPrompt(requestType, fullPrompt, llmProvider) {
    console.log("requestType: " + requestType);
    //console.log("prompt: " + fullPrompt);
    console.log("llmProvider: " + llmProvider);
    
  try {
    const rows = await pool.query('SELECT * FROM chatgpt_responses WHERE request_type = ? AND full_prompt = ? AND llm_provider = ? ORDER BY id DESC LIMIT 1', 
        [requestType, fullPrompt, llmProvider]);

    if (rows.length === 0) {
      return null;
    }

    const row = rows[0];
    const retData = row.ret_data;

    if (!retData || retData.length === 0) {
      return null;
    }

    let parsedData;
    try {
      parsedData = JSON.parse(retData);
    } catch (error) {
      return null;
    }

    parsedData.mock = 'yes';
    //const retString = JSON.stringify(parsedData);

    return parsedData;
  } catch (err) {
    console.error('Error getting mock response:', err);
    return null;
  }
}

async function logRequest(userId, requestType, currentTask, fullPrompt, assistantMessage, retData, llmProvider) {
  const now = new Date();
  const requestDate = now.toISOString().slice(0, 19).replace('T', ' ');
    console.log("logging request");


  try {
    await pool.query('INSERT INTO chatgpt_requests SET ?', {
      user_id: userId,
      request_type: requestType,
      current_task: currentTask,
      full_prompt: fullPrompt,
      assistant_message: assistantMessage,
      ret_data: retData,
      request_date: requestDate,
      llm_provider: llmProvider
    });
  } catch (err) {
    console.error('Error logging request:', err);
  }
}

async function logResponse(userId, requestType, currentTask, fullPrompt, assistantMessage, retData, uid, llmProvider) {
  const now = new Date();
  const requestDate = now.toISOString().slice(0, 19).replace('T', ' ');

  try {
    await pool.query('INSERT INTO chatgpt_responses SET ?', {
      user_id: userId,
      request_type: requestType,
      current_task: currentTask,
      full_prompt: fullPrompt,
      assistant_message: assistantMessage,
      ret_data: retData,
      request_date: requestDate,
      uid: uid,
      llm_provider: llmProvider
    });
  } catch (err) {
    console.error('Error logging response:', err);
  }
}


async function getLatestTaskByRequestType(requestType) {
  try {
    const rows = await pool.query('SELECT * FROM chatgpt_responses WHERE request_type = ? ORDER BY id DESC LIMIT 1', 
        [requestType]);

    if (rows.length === 0) {
      return null;
    }

    const row = rows[0];
    const currentTask = row.current_task;

    if (!currentTask || currentTask.length === 0) {
      return null;
    }

    return currentTask;
  } catch (err) {
    console.error('Error getting latest task by request type:', err);
    return null;
  }
}

async function getLatestResponseByRequestType(requestType) {
  try {
    const rows = await pool.query('SELECT * FROM chatgpt_responses WHERE request_type = ? ORDER BY id DESC LIMIT 1', 
        [requestType]);

    if (rows.length === 0) {
      return null;
    }

    return rows[0];
  } catch (err) {
    console.error('Error getting latest task by request type:', err);
    return null;
  }
}


async function getResponseById(id) {
  try {
    const rows = await pool.query('SELECT * FROM chatgpt_responses WHERE id = ?', [id]);

    if (rows.length === 0) {
      return null;
    }

    const row = rows[0];
    const retData = row.ret_data;

    if (!retData || retData.length === 0) {
      return null;
    }

    let parsedData;
    try {
      parsedData = JSON.parse(retData);
    } catch (error) {
      return null;
    }

    return parsedData;
  } catch (err) {
    console.error('Error getting response by id:', err);
    return null;
  }
}

async function logApplyChanges(allRelevantText, assistantMessage, finalOriginalPaths, finalShortPaths, justFilesLinux, myProjectData, filesSnapshot) {
  try {
    const result = await pool.query('INSERT INTO apply_changes_log SET ?', {
      all_relevant_text: allRelevantText,
      assistant_message: assistantMessage,
      final_original_paths: finalOriginalPaths,
      final_short_paths: finalShortPaths,
      just_files_linux: justFilesLinux,
      my_project_data: myProjectData,
      files_snapshot: filesSnapshot
    });

    console.log('Inserted apply_changes_log with ID:', result.insertId);
    return result.insertId;
  } catch (err) {
    console.error('Error logging apply changes:', err);
    throw err;
  }
}

async function getApplyChangesById(id) {
  try {
    const rows = await pool.query('SELECT * FROM apply_changes_log WHERE id = ?', [id]);

    if (rows.length === 0) {
      return null;
    }

    return rows[0];
  } catch (err) {
    console.error('Error getting apply changes by id:', err);
    return null;
  }
}

async function getResponseByRequestTypeAndUid(requestType, uid) {
  try {
    const rows = await pool.query('SELECT * FROM chatgpt_responses WHERE request_type = ? AND uid = ? ORDER BY id DESC LIMIT 1', 
        [requestType, uid]);

    if (rows.length === 0) {
      return null;
    }

    const row = rows[0];    
    return row;
    
  } catch (err) {
    console.error('Error getting task by request type and uid:', err);
    return null;
  }
}

async function getAllRequestsByUid(uid) {
  try {
    const rows = await pool.query('SELECT * FROM chatgpt_responses WHERE uid = ? ORDER BY id', 
        [uid]);

    if (rows.length === 0) {
      return null;
    }
    return rows;
    
  } catch (err) {
    console.error('Error getting task by request type and uid:', err);
    return null;
  }
}


module.exports = {
  initDatabase,
  logRequest,
  logResponse,
  getMockResponse,
  getMockByFullPrompt,
  getLatestTaskByRequestType,
  getLatestResponseByRequestType,
  getResponseById,
  logApplyChanges,
  getApplyChangesById,
  getResponseByRequestTypeAndUid,
  getAllRequestsByUid
};
