const { ipcRenderer } = require('electron');

console.log("inside scripts.js");

//const worker = new Worker(new URL('./worker.js', import.meta.url), { type: 'module' });

const workerURL = new URL('/worker.bundle.js', window.location.origin);
workerURL.searchParams.set('gpuSettings', JSON.stringify(window.gpuSettings));

let worker;
try {
    worker = new Worker(workerURL, { type: 'module' });
    worker.postMessage({ type: 'init', gpuSettings: window.gpuSettings });
} catch (error) {
    console.error('Error creating worker:', error);
    document.getElementById('status').textContent = 'Error creating worker: ' + error.message;
}

if (worker) {
    worker.onerror = function(error) {
        console.error('Worker error:', error);
        document.getElementById('status').textContent = 'Worker error: ' + error.message;
    };

    worker.onmessageerror = function(error) {
        console.error('Worker message error:', error);
        document.getElementById('status').textContent = 'Worker message error: ' + error.message;
    };
}

let totalModelSize = 0;
let loadedSize = 0;
//const loadModelBtn = document.getElementById('loadModelBtn');
const startRecordingBtn = document.getElementById('startRecordingBtn');
const statusElement = document.getElementById('status');
const transcribeResultElement = document.getElementById('transcribe_result');

let recorder = null;
let audioContext = null;
let chunks = [];
let lastSentTranscription = '';
let analyser = null;

function visualize() {
    const canvas = document.getElementById('visualizer');
    canvas.width = canvas.offsetWidth;
    canvas.height = canvas.offsetHeight;
    const canvasCtx = canvas.getContext('2d');
    const bufferLength = analyser.frequencyBinCount;
    const dataArray = new Uint8Array(bufferLength);

    const drawVisual = () => {
        requestAnimationFrame(drawVisual);
        analyser.getByteTimeDomainData(dataArray);

        canvasCtx.fillStyle = '#0E0E28';
        canvasCtx.fillRect(0, 0, canvas.width, canvas.height);

        canvasCtx.lineWidth = 2;
        canvasCtx.strokeStyle = '#403F85';
        canvasCtx.beginPath();

        const sliceWidth = canvas.width * 1.0 / bufferLength;

        let x = 0;
        for (let i = 0; i < bufferLength; ++i) {
            const v = dataArray[i] / 128.0;
            const y = v * canvas.height / 2;

            if (i === 0) {
                canvasCtx.moveTo(x, y);
            } else {
                canvasCtx.lineTo(x, y);
            }

            x += sliceWidth;
        }

        canvasCtx.lineTo(canvas.width, canvas.height / 2);
        canvasCtx.stroke();
    };

    drawVisual();
}
//loadModelBtn.addEventListener('click', loadModel);
document.addEventListener('DOMContentLoaded', function() {
    console.log("DOM fully loaded and parsed");
    loadModel();
});

startRecordingBtn.addEventListener('click', toggleRecording);

worker.addEventListener('message', handleWorkerMessage);

function loadModel() {
    worker.postMessage({ type: 'load' });
    statusElement.textContent = 'Loading model...';
}

const SILENCE_THRESHOLD = 0.8;
const SILENCE_DURATION = 2000; // 2 second

function detectSilence(audioContext, stream, onSilence) {
  const analyser = audioContext.createAnalyser();
  const source = audioContext.createMediaStreamSource(stream);
  source.connect(analyser);

  analyser.fftSize = 2048;
  const bufferLength = analyser.frequencyBinCount;
  const dataArray = new Uint8Array(bufferLength);

  let silenceStart = null;

  function checkSilence() {
    analyser.getByteTimeDomainData(dataArray);
    let sum = 0;
    for (let i = 0; i < bufferLength; i++) {
      sum += Math.abs(dataArray[i] - 128);
    }
    const average = sum / bufferLength;
    //console.log(`Current average: ${average.toFixed(2)}, Threshold: ${SILENCE_THRESHOLD}`);

    if (average < SILENCE_THRESHOLD) {
      if (!silenceStart) {
        silenceStart = Date.now();
      } else if (Date.now() - silenceStart > SILENCE_DURATION) {
        onSilence();
        return;
      }
    } else {
      silenceStart = null;
    }

    requestAnimationFrame(checkSilence);
  }

  checkSilence();
}

async function toggleRecording() {
    if (recorder && recorder.state === 'recording') {
        recorder.stop();
        startRecordingBtn.classList.remove('recording');
        startRecordingBtn.classList.add('stopped');
        document.getElementById('micIcon').classList.remove('ti-microphone');
        document.getElementById('micIcon').classList.add('ti-microphone-off');
    } else {
        const stream = await navigator.mediaDevices.getUserMedia({ audio: true });
        audioContext = new AudioContext({ sampleRate: 16000 });
        analyser = audioContext.createAnalyser();
        analyser.fftSize = 2048;
        const source = audioContext.createMediaStreamSource(stream);
        source.connect(analyser);
        visualize();
        startRecording(stream);
        startSilenceDetection(stream);
        startRecordingBtn.classList.remove('stopped');
        startRecordingBtn.classList.add('recording');
        document.getElementById('micIcon').classList.remove('ti-microphone-off');
        document.getElementById('micIcon').classList.add('ti-microphone');
    }
}
function startRecording(stream) {
    chunks = [];
    recorder = new MediaRecorder(stream);
    
    recorder.ondataavailable = (e) => {
        if (e.data.size > 0) {
            chunks.push(e.data);
            processAudioChunk();
        }
    };
    
    recorder.onstart = () => {
        startRecordingBtn.innerHTML = '<i id="micIcon" class="ti ti-microphone-2"></i>';
    };
    
    recorder.onstop = () => {
        startRecordingBtn.innerHTML = '<i id="micIcon" class="ti ti-microphone-off"></i>';
    };
    
    recorder.start(1000); // Collect data every second
}

function startSilenceDetection(stream) {
    detectSilence(audioContext, stream, () => {
        console.log('Silence detected');        
        if (recorder && recorder.state === 'recording') {
            console.log("stopping recorder, and restarting");
            sendTranscriptionData(transcribeResultElement.textContent, "silence");
            recorder.stop();
            setTimeout(() => {
                startRecording(stream);
                startSilenceDetection(stream);
            }, 100);
        }
    });
}

async function processAudioChunk() {
    const blob = new Blob(chunks, { type: 'audio/webm' });
    const arrayBuffer = await blob.arrayBuffer();
    const audioBuffer = await audioContext.decodeAudioData(arrayBuffer);
    const audio = audioBuffer.getChannelData(0);
    
    worker.postMessage({ 
        type: 'generate', 
        data: { 
            audio, 
            language: 'en' // You can change this or make it selectable
        } 
    });
}

function sendTranscriptionData(transcriptionText, updateType) {
    let finalText;
    if (Array.isArray(transcriptionText)) {
        finalText = transcriptionText.join(' ');
    } else {
        finalText = transcriptionText;
    }
    
    console.log('Sending transcription data:', finalText);
    ipcRenderer.send('transcription-result', { transcription: finalText, update_type: updateType });
}



function handleWorkerMessage(e) {
    //console.log("handle worker message", e);
    switch (e.data.status) {
        case 'loading':
            statusElement.textContent = `Loading model... ${e.data.data}`;
            break;
        case 'progress':
            updateProgress(e.data);
            break;
        case 'ready':
            statusElement.textContent = 'Model loaded. Ready to record.';
            startRecordingBtn.disabled = false;
            startRecordingBtn.classList.remove('recording', 'stopped');
            document.getElementById('micIcon').classList.remove('ti-microphone-off');
            document.getElementById('micIcon').classList.add('ti-microphone');
            break;
        case 'update':
            transcribeResultElement.textContent = e.data.output;
            sendTranscriptionData(e.data.output, "update");
            break;        
        case 'complete':
            transcribeResultElement.textContent = e.data.output;
            sendTranscriptionData(e.data.output, "complete");
            break;
    }
}
function updateProgress(data) {
    if (data.total) {
        totalModelSize += data.total;
    }
    loadedSize += data.loaded;
    const progress = Math.min(Math.round((loadedSize / totalModelSize) * 100), 100);
    statusElement.textContent = `Loading model... ${progress}%`;
}

// Check if WebGPU is available
if (!navigator.gpu) {
    document.body.innerHTML = '<h1>WebGPU is not supported by this browser :(</h1>';
}

console.log("end of scripts");
ipcRenderer.on('init-transcription', () => {
    // Any initialization code needed when the transcription window is loaded
    console.log('Transcription window initialized');
});

