const { ipcMain } = require('electron');
const path = require('path');


async function generateEmbeddings(docs) {
  let TransformersApi = Function('return import("@xenova/transformers")')();
  const { pipeline } = await TransformersApi;            
  const model = "Xenova/all-MiniLM-L6-v2";
  const generateEmbedding = await pipeline("feature-extraction", model);
  
  const embeddings = await Promise.all(
    docs.map(async (doc) => {
      const filenameWeight = 0.5;
      const descriptionWeight = 0.45;
      const dependenciesWeight = 0.05;
      
      const filenameEmbedding = await generateEmbedding(doc.filename, {
        pooling: "mean",
        normalize: true,
      });
      const descriptionEmbedding = await generateEmbedding(doc.description, {
        pooling: "mean",
        normalize: true,
      });
      const dependenciesEmbedding = await generateEmbedding(doc.dependencies.join("\n"), {
        pooling: "mean",
        normalize: true,
      });
      
      const weightedEmbedding = filenameEmbedding.data.map((val, idx) =>
        val * filenameWeight +
        descriptionEmbedding.data[idx] * descriptionWeight +
        dependenciesEmbedding.data[idx] * dependenciesWeight
      );
      
      return Array.from(weightedEmbedding);
    })
  );

  return embeddings;
}


async function generateDocumentEmbeddings(docs) {
  let TransformersApi = Function('return import("@xenova/transformers")')();
  const { pipeline } = await TransformersApi;            
  const model = "Xenova/all-MiniLM-L6-v2";
  const generateEmbedding = await pipeline("feature-extraction", model);
  
  const embeddings = await Promise.all(
    docs.map(async (doc) => {
      const embeddingNameWeight = 0.3;
      const documentNameWeight = 0.3;
      const markdownWeight = 0.4;
      
      const embeddingNameEmbedding = await generateEmbedding(doc.embedding_name, {
        pooling: "mean",
        normalize: true,
      });
      const documentNameEmbedding = await generateEmbedding(doc.document_name, {
        pooling: "mean",
        normalize: true,
      });
      const markdownEmbedding = await generateEmbedding(doc.markdown, {
        pooling: "mean",
        normalize: true,
      });
      
      const weightedEmbedding = embeddingNameEmbedding.data.map((val, idx) =>
        val * embeddingNameWeight +
        documentNameEmbedding.data[idx] * documentNameWeight +
        markdownEmbedding.data[idx] * markdownWeight
      );
      
      return Array.from(weightedEmbedding);
    })
  );

  return embeddings;
}

async function findMostSimilarDocumentEmbeddings(query, embeddings, topK = 3) {
  const { pipeline } = await import('@xenova/transformers');
  const model = "Xenova/all-MiniLM-L6-v2";
  const generateEmbedding = await pipeline("feature-extraction", model);
  
  const queryEmbedding = await generateEmbedding(query, {
    pooling: "mean",
    normalize: true,
  });
  const similarities = embeddings.map((embedding) =>
    cosineSimilarity(Array.from(queryEmbedding.data), embedding)
  );
  const sortedIndices = similarities
    .map((score, index) => ({ score, index }))
    .sort((a, b) => b.score - a.score)
    .slice(0, topK);

  const mostSimilarIndices = sortedIndices.map((item) => item.index);
  const similarityScores = sortedIndices.map((item) => item.score);

  return { indices: mostSimilarIndices, scores: similarityScores };
}

ipcMain.handle('generateDocumentEmbeddings', async (event, documents) => {
  try {
    const embeddings = await generateDocumentEmbeddings(documents);
    return embeddings;
  } catch (error) {
    console.error('Error generating document embeddings:', error);
    return null;
  }
});

ipcMain.handle('findMostSimilarDocumentEmbeddings', async (event, query, embeddings, topK) => {
  try {
    const { indices, scores } = await findMostSimilarDocumentEmbeddings(query, embeddings, topK);
    return { indices, scores };
  } catch (error) {
    console.error('Error finding most similar document embeddings:', error);
    return null;
  }
});



//STUPID TRANSFORMERS HAVE TO BE HERE???
async function findMostSimilarDocumentIndices(query, embeddings, topK = 3) {
  const { pipeline } = await import('@xenova/transformers');
  const model = "Xenova/all-MiniLM-L6-v2";
  const generateEmbedding = await pipeline("feature-extraction", model);
  const queryEmbedding = await generateEmbedding(query, {
    pooling: "mean",
    normalize: true,
  });
  const similarities = embeddings.map((embedding) =>
    cosineSimilarity(Array.from(queryEmbedding.data), embedding)
  );
  const sortedIndices = similarities
    .map((score, index) => ({ score, index }))
    .sort((a, b) => b.score - a.score)
    .slice(0, topK);

  const mostSimilarIndices = sortedIndices.map((item) => item.index);
  const similarityScores = sortedIndices.map((item) => item.score);

  return { indices: mostSimilarIndices, scores: similarityScores };
}

function cosineSimilarity(a, b) {
  const dotProduct = a.reduce((sum, val, idx) => sum + val * b[idx], 0);
  const normA = Math.sqrt(a.reduce((sum, val) => sum + val * val, 0));
  const normB = Math.sqrt(b.reduce((sum, val) => sum + val * val, 0));
  return dotProduct / (normA * normB);
}

function findTopKIndices(arr, k) {
  const sortedIndices = arr.map((val, idx) => idx).sort((a, b) => arr[b] - arr[a]);
  return sortedIndices.slice(0, k);
}

ipcMain.handle('generateEmbeddings', async (event, documents) => {
  try {
    const embeddings = await generateEmbeddings(documents);
    return embeddings;
  } catch (error) {
    console.error('Error generating embeddings:', error);
    return null;
  }
});

ipcMain.handle('findMostSimilarDocumentIndices', async (event, query, embeddings, topK) => {
  try {
    const { indices, scores } = await findMostSimilarDocumentIndices(query, embeddings, topK);
    return { indices, scores };
  } catch (error) {
    console.error('Error finding most similar documents:', error);
    return null;
  }
});


async function test(){
    const documents = [
      {
        filename: "utility.js",
        description: "This file contains utility functions for string manipulation.",
        dependencies: [
          "file1.js calls function_x from utils.js",
          "file2.js calls function_y from file1.js",
        ],
      },
      {
        filename: "authorize.js",
        description: "This file handles user authentication and authorization.",
        dependencies: [
          "authorize.js calls function_y from utility.js",
          "authorize.js calls function_z from utility.js",
        ],
      },
      {
        filename: "ui_handlers.js",
        description: "This file handles setting up the UI.",
        dependencies: [
          "ui_handlers.js calls get_widgets() from utility.js",
          "ui_handlers.js calls get_layout() from utility.js",
        ],
      },  
    ];    
  const embeddings = await generateEmbeddings(documents);
  const query = "Find files related to building the UI";
  const mostSimilarIndices = await findMostSimilarDocumentIndices(query, embeddings);
  let mostSimilarDocuments = mostSimilarIndices.map((index) => documents[index]);
  console.log("Most similar documents:", mostSimilarDocuments);
}


// Export functions if needed elsewhere
module.exports = {
  generateEmbeddings,
  generateDocumentEmbeddings,
  findMostSimilarDocumentIndices,
  findMostSimilarDocumentEmbeddings,
  cosineSimilarity,
  findTopKIndices,
  test
};